;;; tests.scm -- Demonstration & testing of scsh pg module
;;
;;     Copyright (C) 1999  Eric Marsden
;;     Copyright (C) 2006, 2007  Emilio Lopes
;;
;;     This library is free software; you may redistribute it and/or
;;     modify it under the terms of the GNU Library General Public
;;     License as published by the Free Software Foundation; either
;;     version 2 of the License, or (at your option) any later version.
;;
;;     This library is distributed in the hope that it will be useful,
;;     but WITHOUT ANY WARRANTY; without even the implied warranty of
;;     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;     Library General Public License for more details.
;;
;;     You should have received a copy of the GNU Library General Public
;;     License along with this library; if not, write to the Free
;;     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;
;; Please send suggestions and bug reports to <eclig@gmx.net>.


;; This file runs a limited number of tests on the database code. It
;; assumes you have a database user "postgres" either with no password
;; or with password "postgres"; you can change the username and
;; password to something more appropriate. It uses a special "test"
;; database (which you will have to create first, either using
;; `create_database' from the commandline, or from psql using `CREATE
;; DATABASE'), and should clean up after itself. Provides an example
;; of how to write an executable scsh script, and how to use the
;; module system.
;;
;; * is the postmaster running?
;; * was the postmaster started with the -i commandline option?


(define-structure test
  (export run-tests
          main)
  (open pg
        (modify srfi-19 (prefix srfi-19:))
        srfi-1
        srfi-13
        srfi-78
        formats
        scheme)
  (begin



(define-syntax assert
  (syntax-rules ()
    ((_ op ...)
     (check (op ...) => #t))))

(define-syntax assert-equal?
  (syntax-rules ()
    ((_ op1 op2)
     (check op1 (=> equal?) op2))))

(define-syntax assert-string=?
  (syntax-rules ()
    ((_ op1 op2)
     (check op1 (=> string=?) op2))))

(define-syntax not-exception
  (syntax-rules ()
    ((_ expr)
     (guard (condition
             ((serious-condition? condition) #f)
             (else #t))
       expr
       #t))))

(define-syntax assert-not-exn
  (syntax-rules ()
    ((_ expr)
     (check (not-exception expr) => #t))))

(define (call-with-test-connection proc)
  (let* ((conn (pg:connect "test" "postgres" "postgres"))
         (result (proc conn)))
    (pg:disconnect conn)
    result))

(define (test-general conn)
  (let ((r1 (pg:exec conn "CREATE TABLE pgscmtest (a int, b VARCHAR(4))"))
        (r2 (pg:exec conn "INSERT INTO pgscmtest VALUES (3, 'zae')"))
        (r3 (pg:exec conn "INSERT INTO pgscmtest VALUES (66, 'poiu')"))
        (r4 (pg:exec conn "SELECT * FROM pgscmtest")))
    (assert-string=? (pg:result r1 'status) "CREATE TABLE")
    (assert list? (member "pgscmtest" (pg:tables conn)))
    (assert list? (member "test" (pg:databases conn)))
    (assert-equal? (pg:columns conn "pgscmtest") '("a" "b"))
    (assert integer? (pg:result r2 'oid))
    (assert-string=? (pg:result r2 'status) (format #f "INSERT ~a 1" (pg:result r2 'oid)))
    (assert-string=? (pg:result r4 'status) "SELECT")
    (assert-equal?   (pg:result r4 'attributes) '(("a" 23 4) ("b" 1043 65535)))
    (assert-equal?   (pg:result r4 'tuples) '((3 "zae") (66 "poiu")))
    (assert-equal?   (pg:result r4 'tuple 1) '(66 "poiu"))
    (assert-string=? (pg:result (pg:exec conn "DROP TABLE pgscmtest") 'status) "DROP TABLE")))


(define (test-date conn)
  (pg:exec conn "CREATE TABLE pgscmtest (a timestamp,
                                         b timestamp with time zone,
                                         c time,
                                         d time with time zone,
                                         e date)")
  (pg:exec conn "INSERT INTO pgscmtest VALUES ('1999-01-08 04:05:06',
                                               '1999-01-08 03:05:06+1',
                                               '04:05:06.789',
                                               '04:05:06.789-8',
                                               '1999-01-08')")
  (let* ((res (pg:exec conn "SELECT * FROM pgscmtest"))
         (dates (car (pg:result res 'tuples))))
    (assert-string=? (srfi-19:date->string (first dates) "~w ~j ~H:~M:~S ~Y-~m-~d")
                     "5 008 04:05:06 1999-01-08")
    (assert-string=? (srfi-19:date->string (second dates) "~s") "915761106")
    (assert-string=? (srfi-19:date->string (third dates) "~Y-~m-~d ~S") "1970-01-01 06")
    (assert-string=? (srfi-19:date->string (fourth dates) "~H:~M:~S ~Y-~m-~d") "04:05:06 1970-01-01")
    (assert-string=? (srfi-19:date->string (fifth dates) "~H:~M:~S ~Y-~m-~d") "00:00:00 1999-01-08"))
  (pg:exec conn "DROP TABLE pgscmtest"))

;; test of large-object interface
(define (test-lo-read conn)
  (pg:with-transaction conn
    (let* ((oid (pg:lo-create conn "rw"))
           (fd (pg:lo-open conn oid "rw")))
      (assert integer? oid)
      (assert integer? fd)
      (pg:lo-write conn fd "Hi there mate")
      (pg:lo-lseek conn fd 0 0)         ; SEEK_SET = 0
      (assert zero? (pg:lo-tell conn fd))
      (assert-string=? (pg:lo-read conn fd 7) "Hi ther")
      (pg:lo-close conn fd)
      (pg:lo-unlink conn oid))))

(define (file-as-string fname)
  (call-with-input-file fname
    (lambda (port)
      (string-unfold (lambda (port)
                       (eof-object? (peek-char port)))
                     read-char values port))))

(define (test-lo-import conn)
  (pg:with-transaction conn
    (let* ((import-file "COPYING")
           (export-file "/tmp/__delme__")
           (oid (pg:lo-import conn import-file)))
      (pg:lo-export conn oid export-file)
      (assert-string=? (file-as-string import-file)
                       (file-as-string export-file))
      ;; (delete-file export-file)
      (pg:lo-unlink conn oid))))

(define (run-tests)
  (check-reset!)
  (check-set-mode! 'report-failed)
  (call-with-test-connection
   (lambda (conn)
     (test-general conn)
     (test-date conn)
     (test-lo-read conn)
     (test-lo-import conn)))
  (check-report))

(define (main args)
  (run-tests))

    ))

;;; Local Variables:
;;; eval: (put 'pg:with-transaction 'scheme-indent-function 1)
;;; End:
